import http
import logging
import asyncio

import services.node_service as node_service
import services.job_service as job_service
from quart import (Blueprint, Response, request, json)
from services.training_service import TrainingService

logger = logging.getLogger()

training_service = TrainingService()
p2p_controller = Blueprint('P2P Rest API', __name__)


@p2p_controller.route('/announce', methods=['POST'])
async def node_announce():
    request_json = await request.json
    logger.info('node_announce() method called')
    response = node_service.process_announce_request_and_generate_response(request, request_json)
    return Response(status=http.HTTPStatus.CREATED, response=json.dumps(response), mimetype='application/json')


@p2p_controller.route('/training', methods=['POST'])
async def start_training_in_this_node():
    request_json = await request.json

    logger.info('start_training_in_this_node() method called...')
    job = job_service.find_job_by_trainer_id(request_json['trainerId'])
    if job is None:
        logger.info('start_training_in_this_node() - job not found!')
        return '', http.HTTPStatus.NOT_FOUND

    if -1 < job.get_current_round() < request_json['round'] and not job.is_after_fedavg():
        job.do_federated_average()

    job.update_this_job_with_state_message_body_when_starting_training(request_json)
    asyncio.create_task(training_service.start_training_in_this_node(job))
    return '', http.HTTPStatus.OK


@p2p_controller.route('/training/model_params', methods=['POST'])
async def add_model_params_of_calling_node():
    logger.info('add_model_params_of_calling_node() method called')

    request_json = await request.json

    job = job_service.find_job_by_trainer_id(request_json['trainerId'])
    if job is None:
        logger.info('add_model_params_of_calling_node() - job not found!')
        return '', http.HTTPStatus.NOT_FOUND

    job.add_node_model_params_from_job_state_message_body(request_json, request)
    if node_service.i_am_idle():
        loop = asyncio.get_running_loop()
        asyncio.create_task(training_service.try_to_do_fedavg_for_job_or_prepare_for_doing_it_later(loop, job))
    return '', http.HTTPStatus.OK


@p2p_controller.route('/training/model_params/accuracy', methods=['POST'])
async def obtain_model_params_and_accuracy():
    logger.info('obtain_model_params_and_accuracy() method called')

    request_json = await request.json

    job = job_service.find_job_by_trainer_id(request_json['trainerId'])
    if job is None:
        logger.error('obtain_model_params_and_accuracy() - job not found!')
        return '', http.HTTPStatus.NOT_FOUND

    json_round = request_json['round']
    if not node_service.i_am_idle() or not job.is_current_round(json_round) or not job.is_after_fedavg():
        logger.info('obtain_model_params_and_accuracy() - currently not idle ({}) or received node params for the round'
                    ' I\'m not working on or already did federated average!'.format(node_service.get_status()))
        return '', http.HTTPStatus.BAD_REQUEST

    model_params = request_json['modelParams']
    accuracy = request_json['accuracy']
    node_json = request_json['iAm']
    asyncio.create_task(training_service.process_accuracy_exchange_response(asyncio.get_running_loop(), job, node_json,
                                                                            accuracy, job.get_trainer_accuracy(),
                                                                            model_params))
    return job.to_model_params_accuracy_body(), http.HTTPStatus.OK, {'Content-type': 'application/json'}

