import http
import logging
import asyncio
import configuration.config as config
from quart import (Blueprint, request, render_template)

from configuration.constants.http_verb import HttpVerb
from services.user_interface_service import UserInterfaceService
import services.job_service as job_service
import services.node_service as node_service

logger = logging.getLogger()

user_interface_service = UserInterfaceService()
ui_controller = Blueprint('HTML/CSS/Javascript User Interface', __name__, template_folder='../resources/templates')


@ui_controller.route('/')
async def index():
    logger.info("Call to get HTML main page...")
    nodes = node_service.obtain_nodes()

    asking_for_info_task = None
    if len(nodes) > 0:
        asking_for_info_task = asyncio.create_task(node_service
                                                   .broadcast_message(method=HttpVerb.GET, url_path='/info'))

    node_id = config.NODE_ID
    node_status = node_service.obtain_this_node_status()

    known_nodes = [{'node': node, 'info': None, 'lastStatus': 'UNKNOWN'} for node in nodes]
    available_trainers = job_service.obtain_available_trainers_id_labels()
    current_job_info = job_service.obtain_current_info_per_trainer_for_ui()

    if asking_for_info_task is not None:
        responses = [response['responseJson'] for response in await asking_for_info_task
                     if response['httpStatus'] == http.HTTPStatus.OK]
        for node in known_nodes:
            found_response = [response for response in responses if response['iAm']['id'] == node['node'].node_id]
            if len(found_response) > 0:
                node['info'] = found_response[0]['jobsInfo']
                node['lastStatus'] = found_response[0]['lastNodeStatus']

    return await render_template("index.html",
                                 node_id=node_id,
                                 node_status=node_status,
                                 known_nodes=known_nodes,
                                 available_trainers=available_trainers,
                                 current_job_info=current_job_info)


@ui_controller.route('/info', methods=['GET'])
async def node_info():
    return user_interface_service.obtain_info(), http.HTTPStatus.OK


@ui_controller.route('/training/trigger', methods=['POST'])
async def training_trigger():
    request_json = await request.json

    logger.info('Request POST /training/trigger')

    job = job_service.find_job_by_trainer_id(request_json['trainerId'])
    if job is None:
        return '', http.HTTPStatus.NOT_FOUND

    if node_service.obtain_amount_of_known_nodes() == 0:
        logger.info("This node doesn't know other nodes to start a training session. Ignoring.")
        return '', http.HTTPStatus.BAD_REQUEST

    asyncio.create_task(user_interface_service.tell_network_to_train_and_start_training_in_this_node_too(job))
    return '', http.HTTPStatus.OK


@ui_controller.errorhandler(404)
def page_not_found(error):
    return 'This page does not exist', 404
