import logging
import copy
import functools
import configuration.config as config
import services.node_service as node_service
from trainers.base_trainer import BaseTrainer
from models.training_model_params_model import TrainingModelParamsModel
from models.node_model import NodeModel
from models.federated_learning_model import FederatedLearningModel


class JobModel:
    def __init__(self, trainer=None):
        self.logger = logging.getLogger()
        self.trainer: BaseTrainer = trainer
        self.current_round = -1
        self.federated_learning_params: FederatedLearningModel = trainer.get_default_federated_learning_params()
        self.node_model_params = []
        self.federated_average_already_done = False
        self.model_was_exchanged = False

    def __str__(self):
        return 'Job - Trainer Id: {}, Current round: {}, Amount of accumulated model params: {},' \
               ' Federated Learning params: {}'.format(self.trainer.get_id(), self.current_round, len(self.node_model_params),
                                                       self.federated_learning_params)

    def get_trainer(self):
        return self.trainer

    def get_current_round(self):
        return self.current_round

    def get_federated_learning_params(self):
        return self.federated_learning_params

    def do_federated_average(self):
        received_model_params = []
        for node_model_param in self.node_model_params:
            received_model_params.append(node_model_param.model_params)
        if len(received_model_params) > 0:
            self.logger.info('Finally doing federated average with the weights of {} nodes'.format(len(received_model_params)))
            self.trainer.update_model_params_doing_federated_average(received_model_params)
            self.__clear_data()
            self.federated_average_already_done = True
            self.logger.info("...Federated average done")
        else:
            self.logger.info('No model params from other nodes to do federated average. Ignoring it...')

    def increase_round(self):
        self.current_round += 1

    def to_current_job_state_message_body(self, accuracy=None):
        return {'iAm': node_service.get_this_node().to_request_body(),
                'trainerId': self.trainer.get_id(),
                'modelParams': self.trainer.generate_weight_params_message_body(),
                'round': self.current_round,
                'federatedLearningParams': self.federated_learning_params.to_message_body(),
                'accuracy': accuracy}

    def to_model_params_accuracy_body(self):
        return {'iAm': node_service.get_this_node().to_request_body(),
                'modelParams': self.trainer.generate_weight_params_message_body(),
                'accuracy': self.trainer.get_post_fedavg_accuracy()}

    def to_info_for_ui(self):
        return {'trainer_label': self.trainer.get_label_name(),
                'current_round': self.current_round,
                'federated_learning_params': self.federated_learning_params.__str__(),
                'pre_fedavg_accuracy': self.trainer.get_pre_fedavg_accuracy(),
                'post_fedavg_accuracy': self.trainer.get_post_fedavg_accuracy(),
                'fed_learn_done': 'Yes' if self.federated_average_already_done else 'No',
                'model_was_exchanged': 'Yes' if self.model_was_exchanged else 'No'}

    def is_current_round(self, round):
        return round == self.current_round

    def is_after_fedavg(self):
        return self.federated_average_already_done

    def add_node_model_params_from_job_state_message_body(self, message_body, request):
        node_json = message_body['iAm']
        node_ip = node_json['ip'] if 'ip' in node_json and node_json['ip'] is not None and node_json['ip'] != "" \
            else request.remote_addr
        node = NodeModel(node_id=node_json['id'], ip=node_ip, port=node_json['port'])
        node_service.add_nodes(node)
        self.__add_node_mode_params(message_body['round'],
                                    TrainingModelParamsModel(node,
                                                             self.trainer.process_weight_params_message_body(
                                                                message_body['modelParams'])
                                                             )
                                    )

    def update_this_job_with_state_message_body_when_starting_training(self, message_body):
        if self.current_round == -1:
            self.logger.info("Job {} wasn't started yet in this node. Using model from remote node."
                             .format(self.get_trainer().get_label_name()))
            self.trainer.set_model_params(self.trainer
                                              .process_weight_params_message_body(message_body['modelParams']))

        self.federated_learning_params.process_message_body(message_body['federatedLearningParams'])
        self.current_round = message_body['round']
        self.__clear_data()

    def exchange_our_model_params(self, new_model_params, new_post_fedavg_accuracy):
        self.trainer.set_model_params(new_model_params)
        self.trainer.set_post_fedavg_accuracy(new_post_fedavg_accuracy)
        self.model_was_exchanged = True

    async def train_model_in_a_thread_awaiting(self, loop):
        self.federated_average_already_done = False
        self.model_was_exchanged = False
        federated_learning_params = copy.deepcopy(self.federated_learning_params)
        await loop.run_in_executor(config.APP_THREAD_POOL_EXECUTOR, functools.partial(self.trainer.train_model,
                                                                                      federated_learning_params))
        await loop.run_in_executor(config.APP_THREAD_POOL_EXECUTOR,
                                   functools.partial(self.trainer.update_trainer_model_accuracy_pre_federated_avg,
                                                     federated_learning_params))

    async def calculate_accuracy_of_model_in_a_thread_awaiting(self, loop, model_params):
        federated_learning_params = copy.deepcopy(self.federated_learning_params)
        return await loop.run_in_executor(config.APP_THREAD_POOL_EXECUTOR, functools.partial(
            self.trainer.calculate_model_accuracy, federated_learning_params, model_params))

    async def update_post_fedavg_accuracy_of_trainer_in_a_thread_awaiting(self, loop):
        self.trainer.set_post_fedavg_accuracy(
            await self.calculate_accuracy_of_model_in_a_thread_awaiting(loop, self.trainer.model_params))

    def clear_accuracies_of_trainer(self):
        self.trainer.clear_pre_fedavg_accuracy()
        self.trainer.set_post_fedavg_accuracy(None)

    def get_trainer_accuracy(self):
        return self.trainer.get_post_fedavg_accuracy()

    def __add_node_mode_params(self, round, node_model_params):
        self.logger.info('add_node_mode_params() - Node: {} - Received job round: {} - Current Job information: {}'
                         .format(node_model_params.node, round, self))
        if round < self.current_round:
            self.logger.info('Received a response for an old round. Ignoring it!')
            return
        elif round > self.current_round:
            self.logger.info('Received a response for a greater round than what we are currently. Discarding our data!')
            self.current_round = round
            self.__clear_data()

        if node_model_params.model_params is not None:
            self.node_model_params.append(node_model_params)

    def __clear_data(self):
        self.federated_average_already_done = False
        self.model_was_exchanged = False
        self.node_model_params.clear()
