import asyncio
import http
import logging

from typing import Set

import aiohttp

import configuration.config as config
from configuration.constants.http_verb import HttpVerb

from configuration.constants.node_status import NodeStatus
from models.node_model import NodeModel

node_service_logger = logging.getLogger()

__NODES: Set[NodeModel] = set()
__NODE_STATUS = NodeStatus.IDLE


def obtain_nodes() -> Set[NodeModel]:
    global __NODES

    return __NODES


def clear_nodes():
    global __NODES

    __NODES.clear()


def obtain_amount_of_known_nodes():
    global __NODES

    return len(__NODES)


def __obtain_node_differences(received_nodes):
    global __NODES

    new_nodes = set(received_nodes).difference(__NODES)
    unknown_nodes_to_sender = __NODES.difference(received_nodes)

    return new_nodes, unknown_nodes_to_sender


def add_nodes(received_nodes):
    global __NODES

    if isinstance(received_nodes, NodeModel):
        __NODES.update([received_nodes])
    else:
        __NODES.update(received_nodes)


def remove_node(node):
    global __NODES

    node_service_logger.info("Removing node: {}".format(node))
    if node in __NODES:
        __NODES.remove(node)


def get_status():
    global __NODES

    return __NODE_STATUS


def i_am_idle():
    global __NODE_STATUS

    return __NODE_STATUS == NodeStatus.IDLE


def change_my_status_to(origin, new_node_status):
    global __NODE_STATUS

    __NODE_STATUS = new_node_status
    node_service_logger.info("From '{}', Node status changed to: {}".format(origin, new_node_status))


def obtain_this_node_status():
    global __NODE_STATUS

    return __NODE_STATUS


async def bootstrap_to():
    if config.BOOTSTRAP_PAIRS is None:
        node_service_logger.info('No nodes to bootstrap to!')
        return

    node_service_logger.info('Bootstrapping in {} seconds to give time for the server to start...'.format(config.BOOTSTRAP_WAIT_IN_SECONDS))
    await asyncio.sleep(config.BOOTSTRAP_WAIT_IN_SECONDS)

    for ip_port in config.BOOTSTRAP_PAIRS:
        pair = ip_port.split(':')
        await __announce_to(pair[0].strip(), pair[1].strip())


async def __announce_to(ip, port):
    async with aiohttp.ClientSession() as session:
        new_nodes = await __announce_call(session, ip, port)
        await __announce_to_more_nodes_if_needed(session, new_nodes)


async def __announce_to_more_nodes_if_needed(session, new_nodes):
    if new_nodes is None or new_nodes == set():
        node_service_logger.info('No more unknown peers to announce us to!')
    else:
        for new_node in new_nodes:
            if new_node not in __NODES:
                more_nodes = await __announce_call(session, new_node.ip, new_node.port)
                await __announce_to_more_nodes_if_needed(session, more_nodes)


async def broadcast_message(method, url_path, json_body=None):
    global __NODES

    tasks = []
    async with aiohttp.ClientSession() as session:
        for node in __NODES:
            tasks.append(asyncio.create_task(__do_request_to(session, node, method, url_path, json_body)))

        await asyncio.gather(*tasks)

        responses = []
        for task in tasks:
            if task.exception() is not None or task.result() is None or 'httpStatus' not in task.result()\
                    or task.result()['httpStatus'] == 500:
                if config.PRUNE_UNRESPONSIVE_NODES:
                    remove_node(task.result().node)
                continue

            responses.append(task.result())

        return responses


async def __process_response(node, method, request_url, response):
    try:
        response.raise_for_status()
        node_service_logger.info('Request {} call made to {} - {} successfully.'
                                 .format(method, request_url, node.node_id))
        if response.content_type == 'application/json':
            return {'node': node, 'httpStatus': response.status, 'responseJson': await response.json()}
    except Exception as e:
        node_service_logger.error('Issue doing request {}. It could be a normal response: check Http status code'
                                  ' and request done. {} - {}'.format(method, request_url, node.node_id),
                                  exc_info=e)

    if response is not None:
        return {'node': node, 'httpStatus': response.status}
    else:
        return {'node': node, 'httpStatus': http.HTTPStatus.INTERNAL_SERVER_ERROR}


async def __do_request_to(session, node, method, url_path, json_body):
    request_url = 'http://' + node.ip + ':' + node.port + url_path
    node_service_logger.info('Doing {} call to {}...'.format(method, request_url))

    try:
        if method == HttpVerb.POST:
            async with session.post(request_url, json=json_body, timeout=config.REQUEST_TIMEOUT_IN_SECONDS) as response:
                return await __process_response(node, method, url_path, response)
        elif method == HttpVerb.GET:
            async with session.get(request_url, json=json_body, timeout=config.REQUEST_TIMEOUT_IN_SECONDS) as response:
                return await __process_response(node, method, url_path, response)
    except Exception as e:
        node_service_logger.error('Issue doing request {}. {} - {}'.format(method, request_url, node.node_id),
                                  exc_info=e)
        return {'node': node, 'httpStatus': http.HTTPStatus.INTERNAL_SERVER_ERROR}

    raise NotImplementedError('Http Verb method not implemented!')


async def __announce_call(session, ip, port):
    node_service_logger.info('Announcing ourselves to {}:{}...'.format(ip, port))
    task = asyncio.create_task(__do_request_to(session, NodeModel(node_id=None, ip=ip, port=port), HttpVerb.POST,
                                             '/announce', __generate_announce_request_body()))
    await task

    result = None
    if task.exception() is None and 'httpStatus' in task.result() \
            and task.result()['httpStatus'] == http.HTTPStatus.CREATED:
        result = __process_announce_response(ip, task.result()['responseJson'])
    elif task.exception() is not None:
        node_service_logger.error('The node we tried to announce ourselves ({0}:{1}) is not responding. '
                                  'Exception: {2} Args: {3!r}.'.format(ip, port, type(task.exception()).__name__,
                                                                       task.exception().args), exc_info=task.exception())
    else:
        node_service_logger\
            .error('The node we tried to announce ourselves ({0}:{1}) is not responding with no exception.')

    return result


def __generate_announce_request_body():
    global __NODES

    return {'iAm': get_this_node().to_request_body(),
            'myKnownPeers': [node.to_request_body() for node in __NODES]}


def get_this_node():
    return NodeModel(node_id=config.NODE_ID, ip=config.NODE_IP, port=config.NODE_PORT)


def process_announce_request_and_generate_response(request, request_json):
    if request is None or request_json is None:
        return

    new_nodes = set()
    try:
        known_peers_to_sender = [NodeModel(node_id=node['id'], ip=node['ip'], port=node['port'])
                                 for node in request_json['myKnownPeers']] \
            if request_json['myKnownPeers'] is not None else []

        node_json = request_json['iAm']
        announcer_node = NodeModel(node_id=node_json['id'], ip=node_json['ip'] if 'ip' in node_json else None,
                                   port=node_json['port'])
        if announcer_node.ip is None:
            announcer_node.ip = request.remote_addr
        add_nodes(announcer_node)

        new_nodes, unknown_nodes_to_sender = __obtain_node_differences(known_peers_to_sender)

        result = {'iAm': get_this_node().to_request_body(),
                  'announceYourselfToPeers': []}
        if unknown_nodes_to_sender is not None and unknown_nodes_to_sender != set():
            result['announceYourselfToPeers'] += [node.to_request_body() for node in unknown_nodes_to_sender if
                                                  node.node_id != announcer_node.node_id]
        return result
    finally:
        if new_nodes != set():
            asyncio.create_task(__announce_to_more_nodes_if_needed(new_nodes))


def __process_announce_response(ip, response):
    global __NODES

    node_json = response['iAm']
    new_node = NodeModel(node_id=node_json['id'], ip=ip, port=node_json['port'])
    add_nodes(new_node)

    new_nodes = set()

    announce_to_peers = response['announceYourselfToPeers']

    if announce_to_peers is not None and announce_to_peers != []:
        new_nodes, unknown_nodes_to_sender = __obtain_node_differences(
            [NodeModel(node_id=new_node['id'], ip=new_node['ip'], port=new_node['port'])
             for new_node in announce_to_peers])

    return new_nodes
