import http
import asyncio
import logging
import random

import configuration.config as config
import services.node_service as node_service
from configuration.constants.node_status import NodeStatus
from configuration.constants.http_verb import HttpVerb
from models.job_model import JobModel


class TrainingService:
    def __init__(self):
        self.logger = logging.getLogger()

    async def start_training_in_this_node(self, job: JobModel):
        if not node_service.i_am_idle():
            self.logger.info('Training requested but node is not currently idling')
            return

        self.__cancel_force_federated_average_timer(job)

        self.logger.info('Training requested. Started...')
        node_service.change_my_status_to('start_training_in_this_node_1', NodeStatus.BUSY)
        job.clear_accuracies_of_trainer()

        try:
            loop = asyncio.get_running_loop()
            await job.train_model_in_a_thread_awaiting(loop)

            self.logger.info('Last training requested for {} finished...'.format(job.get_trainer().get_id()))

            self.logger.info('Broadcasting new model params of {} to nodes...'.format(job.get_trainer().get_id()))
            await node_service.broadcast_message(method=HttpVerb.POST, url_path='/training/model_params',
                                                 json_body=job.to_current_job_state_message_body())

            await self.try_to_do_fedavg_for_job_or_prepare_for_doing_it_later(loop, job)
        except Exception as e:
            self.logger.error('Issue at training_service.start_training_in_this_node', exc_info=e)

        node_service.change_my_status_to('start_training_in_this_node_2', NodeStatus.IDLE)

    def __prepare_federated_timer_task(self, job):
        self.__cancel_force_federated_average_timer(job)
        self.logger.info("Preparing timer task to enforce federated learning later for job {}..."
                         .format(job.get_trainer().get_id()))
        config.FORCE_FEDAVG_TIMERS[job.get_trainer().get_id()] = \
            asyncio.create_task(self.__force_federated_average_timer_task(job))

    def __cancel_force_federated_average_timer(self, job):
        if job.get_trainer().get_id() in config.FORCE_FEDAVG_TIMERS:
            self.logger.info("Timer task to enforce federated learning later for job {} cancelled..."
                             .format(job.get_trainer().get_id()))
            timer = config.FORCE_FEDAVG_TIMERS.pop(job.get_trainer().get_id())
            if timer is not None:
                if not timer.cancelled():
                    timer.cancel()

    async def __force_federated_average_timer_task(self, job):
        try:
            wait_time = config.WAIT_TIME_IN_SECONDS_TO_FORCE_FEDAVG + \
                        random.randint(0, config.WAIT_TIME_IN_SECONDS_TO_FORCE_FEDAVG_MARGIN)
            self.logger.info('Wait time for this enforce federated learning timer task will be of {} seconds'
                             .format(wait_time))
            await asyncio.sleep(wait_time)
            self.logger.info("Timer task to enforce federated learning later for job {} triggered!"
                             .format(job.get_trainer().get_id()))
            node_service.change_my_status_to('__force_federated_average_timer_task_1', NodeStatus.BUSY)
            if self.__consider_doing_federated_average(job, force_fedavg=True, called_from_timer=True):
                loop = asyncio.get_running_loop()
                await self.__compute_our_and_external_accuracies_and_exchange_our_model_if_appropiate(job, loop)
        except asyncio.CancelledError:
            node_service.change_my_status_to('__force_federated_average_timer_task_2', NodeStatus.IDLE)
            return

        node_service.change_my_status_to('__force_federated_average_timer_task_3', NodeStatus.IDLE)

    async def __compute_our_and_external_accuracies_and_exchange_our_model_if_appropiate(self, job, loop):
        node_service.change_my_status_to('compute_our_and_external_accuracies_and_exchange_our_model_if_appropiate_1',
                                         NodeStatus.BUSY)

        await job.update_post_fedavg_accuracy_of_trainer_in_a_thread_awaiting(loop)

        node_service.change_my_status_to('compute_our_and_external_accuracies_and_exchange_our_model_if_appropiate_2',
                                         NodeStatus.IDLE)

        current_accuracy = job.get_trainer_accuracy()
        if current_accuracy is None:
            self.logger.error('For some reason, we got to this stage without our accuracy computed...')
            return

        responses = await node_service\
            .broadcast_message(method=HttpVerb.POST, url_path='/training/model_params/accuracy',
                               json_body=job.to_current_job_state_message_body(accuracy=current_accuracy))
        self.logger.info('Got {} responses requesting model params with accuracy. Comparing to ours...'
                         .format(len(responses)))

        await self.__process_accuracy_responses(current_accuracy, job, loop, responses)

    async def __process_accuracy_responses(self, current_accuracy, job, loop, responses):
        for response in responses:
            if 'httpStatus' not in response or response['httpStatus'] != http.HTTPStatus.OK \
                    or 'responseJson' not in response:
                self.logger.info('compute_our_and_external_accuracies_and_exchange_our_model_if_appropiate() - '
                                 'Node responded with no OK status. Ignoring it...')
                continue

            request_json = response['responseJson']
            node_json = request_json['iAm']

            received_accuracy = request_json['accuracy']
            model_params = request_json['modelParams']

            current_accuracy = await self.process_accuracy_exchange_response(loop, job, node_json, received_accuracy,
                                                                             current_accuracy, model_params)

    async def process_accuracy_exchange_response(self, loop, job, node_json, received_accuracy, current_accuracy,
                                                 model_params):
        if received_accuracy is not None \
                and received_accuracy - current_accuracy >= \
                job.get_trainer().get_difference_between_accuracies_to_consider_exchange():
            self.logger.info('Received accuracy from node {} ({}) was greater than ours ({}). Checking it...'
                             .format(node_json, received_accuracy, current_accuracy))
            received_model_params = job.get_trainer().process_weight_params_message_body(model_params)

            accuracy_against_this_node_test_set = await job \
                .calculate_accuracy_of_model_in_a_thread_awaiting(loop, received_model_params)

            if accuracy_against_this_node_test_set - current_accuracy > \
                    job.get_trainer().get_difference_between_accuracies_to_consider_exchange():
                self.logger.info('Model params from node {} were better than ours! Exchanging!'
                                 .format(node_json))
                current_accuracy = accuracy_against_this_node_test_set
                job.exchange_our_model_params(received_model_params, accuracy_against_this_node_test_set)
            else:
                self.logger.info('Model params from node {} was worse or equals than ours. Ignoring it. '
                                 'Accuracy of external model against our valid set: {} - Current accuracy: {}'
                                 .format(node_json, accuracy_against_this_node_test_set, current_accuracy))
        else:
            self.logger.info('Received accuracy from node {} ({}) doesn\'t have much difference (threshold: {}) '
                             'with ours ({}). Ignoring it.'
                             .format(node_json, received_accuracy,
                                     job.get_trainer().get_difference_between_accuracies_to_consider_exchange(),
                                     current_accuracy))

        return current_accuracy

    async def try_to_do_fedavg_for_job_or_prepare_for_doing_it_later(self, loop, job: JobModel):
        try:
            if self.__consider_doing_federated_average(job):
                await self.__compute_our_and_external_accuracies_and_exchange_our_model_if_appropiate(job, loop)
            elif not job.is_after_fedavg():
                self.__prepare_federated_timer_task(job)
        except Exception as e:
            self.logger.error('Issue at training_service.try_to_do_fedavg_for_job_or_prepare_for_doing_it_later',
                              exc_info=e)

    def __consider_doing_federated_average(self, job: JobModel, force_fedavg=False, called_from_timer=False):
        self.logger.info('Considering doing federated average...')
        result = False

        if job.is_after_fedavg():
            self.logger.info('Federated Average already done for this training round! Ignoring.')
            return result

        if (force_fedavg and len(job.node_model_params) > 0)\
                or len(job.node_model_params) == node_service.obtain_amount_of_known_nodes():
            job.do_federated_average()
            result = True
            if not called_from_timer:
                self.__cancel_force_federated_average_timer(job)
        else:
            self.logger.info('...after consideration, we\'ll do federated average later')

        return result
