import logging

import numpy as np
from models.federated_learning_model import FederatedLearningModel


class BaseTrainer:
    def __init__(self, seed=0):
        self.logger = logging.getLogger()
        self.id = None
        self.label_name = None
        self.model_params = None
        self.difference_between_accuracies_to_consider_exchange = 99999.0
        self.pre_federated_avg_accuracy = None
        self.post_federated_avg_accuracy = None
        self.seed = seed
        self.prepare_initial_model_params()

    def prepare_initial_model_params(self):
        raise NotImplementedError('Subclass must implement method')

    def get_default_federated_learning_params(self) -> FederatedLearningModel:
        raise NotImplementedError('Subclass must implement method')

    def train_model(self, federated_learning_params: FederatedLearningModel):
        raise NotImplementedError('Subclass must implement method')

    def calculate_model_accuracy(self, federated_learning_params: FederatedLearningModel, model_params):
        raise NotImplementedError('Subclass must implement method')

    def get_id(self):
        return self.id

    def get_label_name(self):
        return self.label_name

    def set_model_params(self, model_params):
        self.model_params = model_params

    def update_trainer_model_accuracy_pre_federated_avg(self, federated_learning_params:  FederatedLearningModel):
        self.logger.info('Updating pre-federated average accuracy value for trained model...')
        self.pre_federated_avg_accuracy = self.calculate_model_accuracy(federated_learning_params, self.model_params)

    def get_difference_between_accuracies_to_consider_exchange(self):
        return self.difference_between_accuracies_to_consider_exchange

    def get_pre_fedavg_accuracy(self):
        return self.pre_federated_avg_accuracy

    def clear_pre_fedavg_accuracy(self):
        self.pre_federated_avg_accuracy = None

    def get_post_fedavg_accuracy(self):
        return self.post_federated_avg_accuracy

    def set_post_fedavg_accuracy(self, new_accuracy):
        self.post_federated_avg_accuracy = new_accuracy

    def cleanup(self):
        pass

    def update_model_params_doing_federated_average(self, received_model_params):
        model_params_to_fedavg = received_model_params

        if self.model_params is not None:
            model_params_to_fedavg.append(self.generate_weight_params_message_body()['weights'])

        self.model_params = np.stack(np.array(model_params_to_fedavg, dtype=object)).mean(0)

    def generate_weight_params_message_body(self):
        weights_array = None

        if self.model_params is not None:
            weights_array = []
            for i, weights in enumerate(self.model_params):
                weights_array.append(np.array(weights, dtype=object).tolist())

        return {'weights': weights_array}

    def process_weight_params_message_body(self, weight_params_message):
        weights_array = None

        weights_received = weight_params_message['weights'] if 'weights' in weight_params_message else None
        if weights_received is not None:
            weights_array = []
            for weights in weights_received:
                weights_array.append(np.array(weights, dtype=object))

        return weights_array
