import glob
import os
import random
import shutil
import tempfile

import configuration.config as config
from tensorflow.keras import applications
from tensorflow.keras.layers import Dense, Flatten, Conv2D, MaxPool2D
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.preprocessing.image import ImageDataGenerator
from trainers.base_trainer import BaseTrainer
from models.federated_learning_model import FederatedLearningModel


class ChestXRayPneumoniaTrainer(BaseTrainer):
    def __init__(self, seed=0):
        super().__init__(seed)

        self.id = 'CHEST_X_RAY_PNEUMONIA'
        self.label_name = 'Chest X-Ray Pneumonia Dataset'

        self.logger.info('Initializing {} trainer...'.format(self.label_name))

        self.difference_between_accuracies_to_consider_exchange = 0.025

        self.test_data = None

        self.chest_x_ray_dataset_folder = config.DATASETS_PATH + '/chest_xray'
        self.chest_x_ray_temp_folder = config.TMP_PATH + '/chest_xray'
        self.training_data_dirname = '/train'
        self.validation_data_dirname = '/valid'
        self.test_data_dirname = '/test'

        self.temp_folder = None
        self.__create_temp_dataset_folder()

    def prepare_initial_model_params(self):
        model = self.__create_model(self.get_default_federated_learning_params())
        self.model_params = model.get_weights()

    def get_default_federated_learning_params(self):
        return FederatedLearningModel(learning_rate=0.0001, epochs=1, batch_size=2)

    def train_model(self, federated_learning_params: FederatedLearningModel):
        self.test_data = None

        self.__clean_folder(self.training_data_dirname)
        self.__clean_folder(self.validation_data_dirname)
        self.__clean_folder(self.test_data_dirname)

        train_data, validation_data, self.test_data = self.__load_datasets(federated_learning_params)

        model = self.__create_model(federated_learning_params)
        model.set_weights(self.model_params)

        model.fit(x=train_data,
                  steps_per_epoch=10,
                  epochs=federated_learning_params.epochs,
                  validation_data=validation_data,
                  validation_steps=5,
                  verbose=2)

        self.model_params = model.get_weights()

    def cleanup(self):
        self.__clean_temp_dataset_folder()

    def calculate_model_accuracy(self, federated_learning_params: FederatedLearningModel, model_params):
        if self.test_data is None:
            return 0

        model = self.__create_model(federated_learning_params)
        model.set_weights(model_params)
        result = model.evaluate(x=self.test_data, return_dict=True)
        return result['accuracy']

    def __create_model(self, federated_learning_params):
        model = Sequential([
            Conv2D(filters=32, kernel_size=(3, 3), activation='relu', padding='same', input_shape=(224, 224, 3)),
            MaxPool2D(pool_size=(2, 2), strides=2),
            Conv2D(filters=64, kernel_size=(3, 3), activation='relu', padding='same'),
            MaxPool2D(pool_size=(2, 2), strides=2),
            Flatten(),
            Dense(units=2, activation='softmax')
        ])
        model.summary()
        model.compile(optimizer=Adam(learning_rate=federated_learning_params.learning_rate),
                      loss='categorical_crossentropy', metrics=['accuracy'])
        return model

    def __load_datasets(self, federated_learning_params: FederatedLearningModel):
        self.logger.info('Loading CHEST X-RAY IMAGES dataset...')

        image_data_generator = ImageDataGenerator(preprocessing_function=applications.vgg16.preprocess_input)

        origin = self.chest_x_ray_dataset_folder + self.training_data_dirname
        training_destiny = self.temp_folder.name + self.training_data_dirname
        self.__prepare_data(origin, training_destiny, 200)

        # Let's "steal" images from the training data set for the test data set
        origin = self.temp_folder.name + self.training_data_dirname
        test_destiny = self.temp_folder.name + self.test_data_dirname
        self.__prepare_data(origin, test_destiny, round(200*0.20), move_instead=True)

        training_data = image_data_generator.flow_from_directory(directory=training_destiny, target_size=(224, 224),
                                                                 classes=['PNEUMONIA', 'NORMAL'],
                                                                 batch_size=federated_learning_params.batch_size)

        test_data = image_data_generator.flow_from_directory(directory=test_destiny, target_size=(224, 224),
                                                             classes=['PNEUMONIA', 'NORMAL'],
                                                             batch_size=federated_learning_params.batch_size)

        origin = self.chest_x_ray_dataset_folder + self.test_data_dirname
        validation_destiny = self.temp_folder.name + self.validation_data_dirname
        self.__prepare_data(origin, validation_destiny, 134)
        validation_data = image_data_generator.flow_from_directory(directory=validation_destiny, target_size=(224, 224),
                                                                   classes=['PNEUMONIA', 'NORMAL'],
                                                                   batch_size=federated_learning_params.batch_size)

        return training_data, validation_data, test_data

    def __prepare_data(self, dataset_path, tmp_path, sample_size, move_instead=False):
        self.logger.info('Preparing data for {}...'.format(tmp_path))
        os.makedirs(tmp_path + '/NORMAL')
        os.makedirs(tmp_path + '/PNEUMONIA')

        if move_instead:
            self.__move_images_from_dataset(dataset_path, tmp_path, ['NORMAL', 'PNEUMONIA'], sample_size)
        else:
            self.__copy_images_from_dataset(dataset_path, tmp_path, ['NORMAL', 'PNEUMONIA'], sample_size)

    def __clean_folder(self, dataset_suffix):
        dataset_path = self.temp_folder.name + dataset_suffix
        self.logger.info('Cleaning the data folder: {}'.format(dataset_path))
        shutil.rmtree(dataset_path, ignore_errors=True)

    def __create_temp_dataset_folder(self):
        if os.path.isdir(self.chest_x_ray_temp_folder) is False:
            self.logger.info('Temporary dataset folder {} doesn\'t exist, creating it'
                             .format(self.chest_x_ray_temp_folder))
            os.mkdir(self.chest_x_ray_temp_folder)

        self.temp_folder = tempfile.TemporaryDirectory(dir=self.chest_x_ray_temp_folder)
        self.logger.info('Temporary folder for training: {}'.format(self.temp_folder.name))

    def __clean_temp_dataset_folder(self):
        self.logger.info('Deleting content of temporary folder: {}'.format(self.temp_folder.name))
        self.temp_folder.cleanup()
        if os.path.isdir(self.chest_x_ray_temp_folder):
            path, dirs, files = next(os.walk(self.chest_x_ray_temp_folder))
            if len(dirs) == 0:
                self.logger.info('Deleting temporary dataset folder: {}'.format(self.chest_x_ray_temp_folder))
                os.rmdir(self.chest_x_ray_temp_folder)

    def __copy_images_from_dataset(self, dataset_path, destiny_path, classes, samples_size, pattern='*'):
        for a_class in classes:
            for random_file in random.sample(glob.glob(dataset_path + '/' + a_class + '/' + pattern), samples_size):
                shutil.copy(random_file, destiny_path + '/' + a_class)

    def __move_images_from_dataset(self, dataset_path, destiny_path, classes, samples_size, pattern='*'):
        for a_class in classes:
            for random_file in random.sample(glob.glob(dataset_path + '/' + a_class + '/' + pattern), samples_size):
                shutil.move(random_file, destiny_path + '/' + a_class)
