import numpy

from tensorflow.keras.datasets import cifar10
from tensorflow.keras.utils import to_categorical
from tensorflow.keras.layers import Dense, Flatten, Conv2D
from tensorflow.keras.models import Sequential
from tensorflow.python.keras.layers import Dropout, MaxPooling2D
from tensorflow.python.keras.optimizer_v2.gradient_descent import SGD
from trainers.base_trainer import BaseTrainer
from models.federated_learning_model import FederatedLearningModel

# CIFAR-10 CNN Trainer based on this Kaggle notebook from Justin Güse:
# https://www.kaggle.com/guesejustin/object-recognition-mlp-cnn-efficientnet-on-cifar


class Cifar10Trainer(BaseTrainer):
    def __init__(self, seed=0):
        super().__init__(seed)

        self.id = 'CIFAR-10'
        self.label_name = 'CIFAR-10 Dataset'

        self.logger.info('Initializing {} trainer...'.format(self.label_name))

        self.difference_between_accuracies_to_consider_exchange = 0.05

        self.test_data = None

    def prepare_initial_model_params(self):
        model = self.__create_model(self.get_default_federated_learning_params())
        self.model_params = model.get_weights()

    def get_default_federated_learning_params(self):
        return FederatedLearningModel(learning_rate=0.0001, epochs=1, batch_size=2)

    def train_model(self, federated_learning_params: FederatedLearningModel):
        self.test_data = None
        train_batches, validation_data, self.test_data = self.__load_datasets()

        model = self.__create_model(federated_learning_params)
        model.set_weights(self.model_params)

        model.fit(x=train_batches[0], y=train_batches[1],
                  validation_data=validation_data,
                  epochs=federated_learning_params.epochs,
                  batch_size=federated_learning_params.batch_size)

        self.model_params = model.get_weights()

    def calculate_model_accuracy(self, federated_learning_params: FederatedLearningModel, model_params):
        if self.test_data is None:
            return 0

        model = self.__create_model(federated_learning_params)
        model.set_weights(model_params)
        result = model.evaluate(x=self.test_data[0], y=self.test_data[1], return_dict=True)
        return result['accuracy']

    def __create_model(self, federated_learning_params):
        model = Sequential()
        model.add(Conv2D(filters=16, kernel_size=2, padding='same', activation='relu',
                         input_shape=(32, 32, 3)))
        model.add(MaxPooling2D(pool_size=2))
        model.add(Conv2D(filters=32, kernel_size=2, padding='same', activation='relu'))
        model.add(MaxPooling2D(pool_size=2))
        model.add(Conv2D(filters=64, kernel_size=2, padding='same', activation='relu'))
        model.add(MaxPooling2D(pool_size=2))
        model.add(Dropout(0.3, seed=self.seed))
        model.add(Flatten())
        model.add(Dense(500, activation='relu'))
        model.add(Dropout(0.4, seed=self.seed))
        model.add(Dense(10, activation='softmax'))

        model.summary()

        model.compile(loss='categorical_crossentropy', optimizer=SGD(lr=federated_learning_params.learning_rate),
                      metrics=['accuracy'])

        return model

    def __load_datasets(self):
        self.logger.info('Loading CIFAR-10 dataset...')

        (x_train, y_train), (x_valid, y_valid) = cifar10.load_data()
        num_classes = len(numpy.unique(y_train))

        # Let's keep random 50% of validation data each time
        idx = numpy.random.choice(numpy.arange(len(x_valid)), round(len(x_valid) * 0.50), replace=False)
        x_valid = x_valid[idx]
        y_valid = y_valid[idx]

        x_valid = x_valid.astype('float32')
        x_valid /= 255
        y_valid = to_categorical(y_valid, num_classes)

        x_train = x_train.astype('float32')
        x_train /= 255
        y_train = to_categorical(y_train, num_classes)

        # Let's "steal" data from original training data
        half_pos = round(len(x_train)*0.5)
        x_test = x_train[half_pos:]
        y_test = y_train[half_pos:]
        idx = numpy.random.choice(numpy.arange(len(x_test)), round(len(x_test) * 0.20), replace=False)
        x_test = x_test[idx]
        y_test = y_test[idx]

        # Keep random 50% of the initial 5000 samples from original training data to "simulate" "new" training data
        # each round
        x_train = x_train[:half_pos]
        y_train = y_train[:half_pos]
        idx = numpy.random.choice(numpy.arange(len(x_train)), round(len(x_train) * 0.50), replace=False)
        x_train = x_train[idx]
        y_train = y_train[idx]

        return (x_train, y_train), (x_valid, y_valid), (x_test, y_test)
