import random

import torch
import numpy as np
from PIL import Image
from fastai.data.external import untar_data, URLs
from fastai.data.load import DataLoader
from fastai.torch_core import to_np, tensor

from models.federated_learning_model import FederatedLearningModel
from trainers.base_trainer import BaseTrainer


# MNIST Trainer based on the tutorial from the O'Reilly book:
# Deep Learning for Coders with fastai and PyTorch by Jeremy Howard, Sylvain Gugger
# https://www.oreilly.com/library/view/deep-learning-for/9781492045519/ch04.html


class MnistTrainer(BaseTrainer):
    def __init__(self, seed=0):
        super().__init__(seed)

        self.id = 'MNIST'
        self.label_name = 'MNIST Dataset'

        self.logger.info('Initializing {} trainer...'.format(self.label_name))

        self.difference_between_accuracies_to_consider_exchange = 0.05
        self.ACCURACY_THRESHOLD = 0.5

        self.test_dataloader = None

    def prepare_initial_model_params(self):
        weights = torch.randn((28 * 28, 1), dtype=torch.float, requires_grad=True)
        bias = torch.randn(1, dtype=torch.float, requires_grad=True)
        self.model_params = (weights, bias)

    def get_default_federated_learning_params(self):
        return FederatedLearningModel(learning_rate=0.00001, epochs=2, batch_size=64)

    def train_model(self, federated_learning_params: FederatedLearningModel):
        self.test_dataloader = None

        training_dataset, validation_dataset, test_dataset = self.__load_datasets()

        training_dataloader = DataLoader(training_dataset, batch_size=federated_learning_params.batch_size)
        validation_dataloader = DataLoader(validation_dataset, batch_size=federated_learning_params.batch_size)
        self.test_dataloader = DataLoader(test_dataset, batch_size=federated_learning_params.batch_size)

        for epoch in range(federated_learning_params.epochs):
            self.__train_epoch(federated_learning_params, training_dataloader)
            self.logger.info('Accuracy of model trained at epoch {}: {}'
                             .format(epoch + 1, self.__validate_epoch(self.model_params, validation_dataloader)))

    def update_model_params_doing_federated_average(self, received_model_params):
        received_weights = []
        received_biases = []

        if self.model_params is not None and self.model_params[0] is not None and self.model_params[1] is not None:
            numpy_params = to_np(self.model_params)
            received_weights.append(torch.tensor(np.array(numpy_params[0].tolist()), dtype=torch.float,
                                                 requires_grad=True))
            received_biases.append(torch.tensor(np.array(numpy_params[1].tolist()), dtype=torch.float,
                                                requires_grad=True))

        for received_params in received_model_params:
            received_weights.append(received_params[0])
            received_biases.append(received_params[1])

        new_weights = torch.stack(received_weights).mean(0).clone().detach().requires_grad_(True)
        new_bias = torch.stack(received_biases).mean(0).clone().detach().requires_grad_(True)
        self.model_params = new_weights, new_bias

    def generate_weight_params_message_body(self):
        numpy_params = to_np(self.model_params)
        return {'weights': numpy_params[0].tolist(), 'bias': numpy_params[1].tolist()}

    def process_weight_params_message_body(self, weight_params_message):
        weights = torch.tensor(np.array(weight_params_message['weights']), dtype=torch.float, requires_grad=True)
        bias = torch.tensor(np.array(weight_params_message['bias']), dtype=torch.float, requires_grad=True)
        return weights, bias

    def calculate_model_accuracy(self, federated_learning_params: FederatedLearningModel, model_params):
        if self.test_dataloader is None:
            return 0

        return self.__validate_epoch(model_params, self.test_dataloader)

    def __train_epoch(self, federated_learning_params: FederatedLearningModel, dataset_dataloader):
        for train_data, train_labels in dataset_dataloader:
            self.__calculate_gradients(train_data, train_labels)
            for model_param in self.model_params:
                model_param.data -= model_param.grad * federated_learning_params.learning_rate
                model_param.grad.zero_()

    def __validate_epoch(self, model_params, dataset_dataloader):
        accuracies = [self.__accuracy(self.__linear_model(train_data, weights=model_params[0], bias=model_params[1]),
                                      train_labels) for train_data, train_labels in dataset_dataloader]
        return round(torch.stack(accuracies).mean().item(), 4)

    def __calculate_gradients(self, train_data, train_labels):
        predictions = self.__linear_model(train_data, self.model_params[0], self.model_params[1])
        loss = self.__mnist_loss(predictions, train_labels)
        loss.backward()

    def __mnist_loss(self, predictions, targets):
        predictions = predictions.sigmoid()
        return torch.where(targets == 1, 1 - predictions, predictions).mean()

    def __linear_model(self, matrix, weights, bias):
        return matrix @ weights + bias

    def __accuracy(self, train_data, train_labels):
        predictions = train_data.sigmoid()
        corrections = (predictions > self.ACCURACY_THRESHOLD) == train_labels
        return corrections.float().mean()

    def __load_datasets(self):
        self.logger.info('Loading dataset MNIST_SAMPLE...')

        path = untar_data(URLs.MNIST_SAMPLE)
        self.logger.info('MNIST dataset at: {}'.format(path))

        training_test_dataset, amount_threes, amount_sevens = self.__prepare_set(path, 'train', 30, 70)
        validation_dataset, _, _ = self.__prepare_set(path, 'valid', 10, 30)

        # Let's "steal" part of the training data as test dataset
        threes_part = round(amount_threes * 0.80)
        sevens_part = round(amount_sevens * 0.80)

        training_dataset = training_test_dataset[:threes_part] + \
                           training_test_dataset[amount_threes:(amount_threes + sevens_part)]
        test_dataset = training_test_dataset[threes_part:amount_threes] + \
                       training_test_dataset[(amount_threes + sevens_part):]

        return training_dataset, validation_dataset, test_dataset

    def __prepare_set(self, path, type_path, random_min, random_max):
        paths_for_threes = random.sample((path / type_path / '3').ls().sorted(),
                                         int(random.uniform(random_min, random_max)))
        paths_for_sevens = random.sample((path / type_path / '7').ls().sorted(),
                                         int(random.uniform(random_min, random_max)))

        three_tensors = [tensor(Image.open(image_path)) for image_path in paths_for_threes]
        seven_tensors = [tensor(Image.open(image_path)) for image_path in paths_for_sevens]

        stacked_threes = torch.stack(three_tensors).float() / 255
        stacked_sevens = torch.stack(seven_tensors).float() / 255

        images = torch.cat([stacked_threes, stacked_sevens]).view(-1, 28 * 28)
        labels = tensor([1] * len(three_tensors) + [0] * len(seven_tensors)).unsqueeze(1)
        return list(zip(images, labels)), len(stacked_threes), len(stacked_sevens)
